# CSS Workaround - Wrong order after render

### Background
When the renderToString() function was used for server-side rendering,
it occasionally resulted in the incorrect ordering of elements in the DOM.
The tabs component, for instance, displayed this issue prominently,
causing a disruptive flashing effect as the client-side JavaScript took a moment
to reorder the elements correctly.
\newline

renderToString result from server:

```
<div ... class="sc-sdx-tabs tablist">
    ...
    <button ... >
    <!--t.1.5.4.0-->
    Tab 6
    </button>
    <button ... >
    <!--t.1.7.4.0-->
    Tab 4
    </button>
    <button ... >
    <!--t.1.9.4.0-->
    Tab 2
    </button>
    ...
</div>
``` 


From our findings, the most effective workaround for this issue was the implementation of CSS.
We utilized the CSS Grid model to control the order of elements visually.
This approach allowed us to manipulate the display order of the elements
without altering their actual DOM order.

### Implementation Approach

The viable solution involves generating the `grid-template-areas` property by having prior knowledge of the IDs assigned to elements.
Each element can then be assigned a `grid-area` based on its corresponding ID.

To ensure the correct order for `grid-template-areas`,
a potential solution involves pre-knowing the IDs assigned to the tabs.

In this example the assigned ID for tab items is in the format "tab1", "tab2" etc.
The prefix `tab` is a constant part of the ID, while the numerical value is determined dynamically.

The following steps were taken to achieve the correct order of the tabs.


#### 1. Assining IDs to tabs elements in HTML

`
<sdx-tabs-item id="label="Tab 1">...</sdx-tabs-item>
`

etc.



#### 2. Generating grid-template-areas

Since it is possible to get the number of child elements (tab items), the areas can be generated independently of the number of elements.
As established earlier, the prefix for all Tab IDs is "tab," and it is hard-coded.
By iterating through the array of tab elements, the loop index was appended to the prefix
and pushed into the array of `grid-template-areas`.

```
private getTabIds(tabsItemEls: HTMLSdxTabsItemElement[]) {
// return tabsItemEls.map(element => element.id); // WONT work

    let gridAreas = [];
    for (let i = 1; i <= tabsItemEls.length; i++) {
        gridAreas.push('tab' + i)
    }
    return gridAreas;
}

public render() {

    const { 
        tabsItemEls, 
        selectedTabsItemEl 
        } = this.state.get()

    ...
    const elementIdsForGrid = this.getTabIds(tabsItemEls)
    ...
}
```



#### 3. Adding grid-template-area to the parent of tab items

The CSS Grid layout should be added to the parent container of the tabs component.
By defining grid areas, it is possible to control the visual placement of child elements.

```
return (
...
<div
	class="tablist"
	role="tablist"
	aria-label={this.srHint}
	ref={(el) => (this.tablistEl = el)}
	style={{
	  display: `grid`,
	  gridTemplateAreas: `'${elementIdsForGrid.join(' ')}'`,
	}}
	>
	...      
)  
```


#### 4. Assigning grid-area to each tab item

The style attribute should be set for each child, assigning the `grid-area` based on their respective IDs.

```
const Tag = tabsItemEl.href ? "a" : "button"

return (
<Tag
class={this.getClickableElClassNames(tabsItemEl)}
@\bh@style={{gridArea: tabsItemEl.id}}@\eh@
...

    >
        {tabsItemEl.label}
    </Tag>
)
```


### Results
This workaround successfully mitigated the visual disturbance caused by the incorrect ordering of elements.
The tabs component now maintains a consistent visual order from the moment it loads,
significantly enhancing the user experience.

While this approach does not solve the root cause of the ordering issue,
it provides a stable and effective temporary solution.

The corrected `renderToString()` result from server:

```
<div ... class="sc-sdx-tabs tablist" style="display: grid; grid-template-areas: 'tab1 tab2 tab3 tab4 tab5 tab6';">
    ...
    <button ... style="grid-area: tab6;">
    <!--t.1.5.4.0-->
    Tab 6
    </button>
    <button ... style="grid-area: tab4;">
    <!--t.1.7.4.0-->
    Tab 4
    </button>
    <button ... style="grid-area: tab2;">
    <!--t.1.9.4.0-->
    Tab 2
    </button>

</div>
```
